/*----------------------------------------------------------------------
  File    : wchull.c
  Contents: Windows program to construct and visualize a convex hull
  Author  : Christian Borgelt
  History : 2001.08.07 file created from file wsom.c
            2011.12.09 adapted to utility module random
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include <time.h>
#include <windows.h>
#include "wchull.rh"
#include "random.h"
#ifndef CH_DRAW
#define CH_DRAW
#endif
#include "chull.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#ifndef M_PI                    /* if not defined (no Posix source) */
#define M_PI      3.14159265358979323846  /* \pi */
#endif

#define PRGNAME  "wchull"       /* main window title */

/* --- abbreviations --- */
#define HINST             HINSTANCE
#define MKIRSC            MAKEINTRESOURCE
#define SDIMSG            SendDlgItemMessage
#define WINSTYLE          WS_OVERLAPPEDWINDOW
#define PATH_MAX          _MAX_PATH

/* --- edit box functions --- */
#define edt_set(h,i,s)    SetDlgItemText(h, i, s)
#define edt_get(h,i,s,n)  GetDlgItemText(h, i, s, n)
#define int_set(h,i,n)    SetDlgItemInt(h, i, n, 0)
#define int_get(h,i)      GetDlgItemInt(h, i, NULL, 0)

/* --- combo box functions --- */
#define cmb_add(h,i,s)    SDIMSG(h, i, CB_ADDSTRING,    0, (LPARAM)(s))
#define cmb_rem(h,i,k)    SDIMSG(h, i, CB_DELETESTRING, k, 0)
#define cmb_set(h,i,k)    SDIMSG(h, i, CB_SETCURSEL,    k, 0)
#define cmb_get(h,i)      SDIMSG(h, i, CB_GETCURSEL,    0, 0)

/* --- extensions --- */
#define INITXEXT     320        /* initial x-extension of window */
#define INITYEXT     320        /* initial y-extension of window */

/* --- movement modes --- */
#define ROTATE         1        /* change heading and pitch */
#define YAW            2        /* change heading */
#define PITCH          3        /* change pitch angle */
#define ROLL           4        /* change roll  angle */
#define ZOOM           5        /* change distance to the origin */

/* --- error codes --- */
#define E_NONE         0        /* no error */
#define E_NOMEM      (-1)       /* not enough memory */
#define E_FOPEN      (-2)       /* cannot open file */
#define E_FREAD      (-3)       /* read error on file */
#define E_FWRITE     (-4)       /* write error on file */
#define E_INIT       (-5)       /* initialization failed */
#define E_UNKNOWN    (-6)       /* unkown error */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
const char *errmsgs[] = {       /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file\n%s",
  /* E_FREAD    -3 */  "read error on file\n%s",
  /* E_FWRITE   -4 */  "write error on file\n%s",
  /* E_INIT     -5 */  "initialization failed",
  /* E_UNKNOWN  -7 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static HINST   hinst;           /* instance handle */
static HWND    hmain;           /* handle of main window */
static HDC     hdc_win;         /* device context of window */
static HDC     hdc_buf;         /* device context of drawing buffer */
static HBITMAP buffer = NULL;   /* buffer for drawing */
static char    text[_MAX_PATH +1024];

/* --- pens and brushes --- */
static HPEN    pen_black;       /* black : grid of map */
static HPEN    pen_grey;        /* grey  : shape of pattern region */
static HBRUSH  br_white;        /* white : background */

/* --- neural network information --- */
static struct {                 /* -- parameters -- */
  int    ptcnt;                 /* number of points */
  int    region;                /* region of points */
  int    alg;                   /* construction algorithm */
  int    visib;                 /* visibility of edges */
} params = { 100, CH_SURFACE, CH_SORTED, CH_SOLID };

static CHULL  *chull  = NULL;   /* convex hull structure */
static int    state   = 0;      /* construction state */
static PROJ3D *proj3d = NULL;   /* 3D to 2D projection */
static int    mode    = 0;      /* movement mode */
static int    mx = 0, my = 0;   /* old mouse coordinates */
static double xoff, yoff;       /* coordinate offsets */
static double xscl, yscl;       /* scaling factors */
static HDC    hdc_draw;         /* device context for drawing */

/*----------------------------------------------------------------------
  Error Functions
----------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print error message */
  va_list     args;             /* list of variable arguments */
  const char  *msg;             /* error message */

  if ((code > 0) || (code < E_UNKNOWN))
    code = E_UNKNOWN;           /* check error code */
  va_start(args, code);         /* get variable arguments */
  msg = errmsgs[-code];         /* and error message text */
  if (!msg) msg = errmsgs[-E_UNKNOWN];
  vsprintf(text, msg, args);    /* format error message */
  MessageBox(hmain, text, PRGNAME, MB_OK|MB_ICONSTOP);
  va_end(args);                 /* show alert dialog box */
}  /* error() */

/*----------------------------------------------------------------------
  Drawing Functions
----------------------------------------------------------------------*/

static void drawpt (double x, double y, int op)
{                               /* --- move to/draw a line to a point */
  int cx, cy;                   /* coordinates of current point */

  cx = (int)(xoff +x *xscl);    /* transform the coordinates */
  cy = (int)(yoff -y *xscl);    /* w.r.t. the window size */
  if (op & 1)                   /* if to draw a line */
    LineTo(hdc_draw, cx, cy);   /* to the new point */
  if (op & 2) {                 /* if to mark the new point */
    MoveToEx(hdc_draw, cx-1, cy, NULL); LineTo(hdc_draw, cx+1, cy);
    MoveToEx(hdc_draw, cx, cy-1, NULL); LineTo(hdc_draw, cx, cy+1);
  }
  MoveToEx(hdc_draw, cx, cy, NULL);
}  /* drawpt() */               /* go to new point */

/*--------------------------------------------------------------------*/

static void redraw (HDC hdc)
{                               /* --- Actions > Redraw */
  static int owd = 0, oht = 0;  /* old window extensions */
  static int opc = 0;           /* old pattern counter */
  RECT       rc;                /* rectangle for drawing */
  int        wd, ht;            /* width and height of window */

  GetClientRect(hmain, &rc);    /* get the window's client rectangle */
  wd = rc.right  -rc.left +1;   /* and compute its width and height */
  ht = rc.bottom -rc.top  +1;   /* in pixels (it may have changed) */
  hdc_draw = hdc_buf;           /* default: draw into the buffer */
  if ((wd != owd) || (ht != oht)/* if the window extensions changed, */
  ||  !buffer) {                /* a new buffer has to be allocated */
    if (buffer) DeleteObject(buffer);
    buffer = CreateCompatibleBitmap(hdc_win, wd, ht);
    if (buffer) SelectObject(hdc_buf, buffer);
    else hdc_draw = hdc_win;    /* if no buffer can be allocated, */
  }                             /* draw directly into the window */
  owd = wd; oht = ht;           /* note the (new) window dimensions */
  xoff = 0.5*wd; yoff = 0.5*ht;
  xscl = 0.5*(wd-1);            /* compute the coordinate offsets */
  yscl = 0.5*(ht-1);            /* and the scaling factors */
  FillRect(hdc_draw, &rc, br_white);
  SelectObject(hdc_draw, pen_black);
  if (chull)                    /* draw the convex hull */
  ch_draw(chull, params.visib, proj3d, drawpt);
  if (hdc_draw == hdc_buf) {    /* if it was drawn to the buffer */
    if (!hdc) hdc = hdc_win;    /* get device context of window */
    BitBlt(hdc, 0, 0, wd-1, ht-1, hdc_buf, 0, 0, SRCCOPY);
  }                             /* copy the buffer to the window */
}  /* redraw() */

/*--------------------------------------------------------------------*/

static void mouse (int x, int y)
{                               /* --- evaluate mouse movement */
  int    dx, dy;                /* pixels mouse was moved */
  double sf;                    /* scaling factor */

  dx = x -mx;                   /* get the number of pixels */
  dy = y -my;                   /* the mouse was moved */
  sf = 75.0 /xscl;              /* compute the scaling factor */
  switch (mode) {               /* according to the movement mode */
    case ROTATE: if ((dx == 0) && (dy == 0)) return;
                 p3d_restore(proj3d);
                 p3d_pitch(proj3d, dy*sf*(M_PI/180), P3D_BOTH);
                 p3d_yaw  (proj3d, dx*sf*(M_PI/180), P3D_BOTH); break;
    case YAW:    if (dx == 0) return;
                 p3d_restore(proj3d);
                 p3d_yaw  (proj3d, dx*sf*(M_PI/180), P3D_BOTH); break;
    case PITCH:  if (dy == 0) return;
                 p3d_restore(proj3d);
                 p3d_pitch(proj3d, dy*sf*(M_PI/180), P3D_BOTH); break;
    case ROLL:   if (dx == 0) return;
                 p3d_restore(proj3d);
                 p3d_roll (proj3d, dx*sf*(M_PI/180), P3D_BOTH); break;
    case ZOOM:   if (dy == 0) return;
                 p3d_restore(proj3d);
                 p3d_move (proj3d, 0, dy*0.02, 0);              break;
  }                             /* change the viewing parameters */
}  /* mouse() */

/*--------------------------------------------------------------------*/

static void mn_enable (int state)
{                               /* --- enable/disable menu items */
  HMENU menu;                   /* handle of window menu */
  UINT  flag;                   /* enable/disable flag */

  menu = GetMenu(hmain);        /* get menu handle */
  flag = (state < 2) ? MF_ENABLED : MF_GRAYED;
  EnableMenuItem(menu, MI_NEXT,  flag);
  EnableMenuItem(menu, MI_BUILD, flag);
}  /* mn_enable() */

/*----------------------------------------------------------------------
  Convex Hull Functions
----------------------------------------------------------------------*/

static void init (void)
{                               /* --- initialize set of points */
  int    i;                     /* loop variable */
  double x, y, z;               /* coordinates of random points */

  if (chull) ch_delete(chull);  /* delete existing structure */
  chull = ch_create();          /* create a new convex hull structure */
  if (!chull) { error(E_NOMEM); return; }
  for (i = params.ptcnt; --i >= 0; ) {   /* generate random points */
    ch_rand(params.region, drand, &x, &y, &z);
    if (ch_point(chull, x, y, z) != 0) {
      error(E_NOMEM); ch_delete(chull); chull = NULL; return; }
  }                             /* add the points to the structure */
  state = 0;                    /* and the construction state */
  mn_enable(state);             /* enable/disable menu items */
}  /* init() */

/*--------------------------------------------------------------------*/

static void next (void)
{                               /* --- add next point */
  int r;                        /* result of ch_next() */

  if (state > 1) return;        /* check the construction state */
  if (state > 0)                /* if construction is already started */
    r = ch_next(chull);         /* process the next point */
  else                          /* if construction is not started, */
    r = ch_start(chull, params.alg);     /* start the construction */
  if (r < 0) error(E_NOMEM);    /* check for an error and */
  state = (r != 0) ? 2 : 1;     /* update the construction state */
  mn_enable(state);             /* enable/disable menu items */
}  /* next() */

/*--------------------------------------------------------------------*/

static void build (void)
{                               /* --- build convex hull */
  static HCURSOR hcwait;        /* buffer for hourglass cursor */
  char    buf[64];              /* buffer for timer output */
  HCURSOR hcnorm;               /* normal cursor */
  clock_t t;                    /* starting time */
  int     r;                    /* result of ch_build() */
  double  time;                 /* processing time */

  if (state > 1) return;        /* check the construction state */
  if (!hcwait) hcwait = LoadCursor(NULL, IDC_WAIT);
  hcnorm = SetCursor(hcwait);   /* show hourglass while computing */
  t = clock();                  /* get the starting time */
  r = ch_build(chull, params.alg); /* build the complex hull */
  t = clock() -t;               /* compute the running time */
  state = 2;                    /* update the construction state */
  mn_enable(state);             /* enable/disable menu items */
  SetCursor(hcnorm);            /* restore normal cursor */
  if (r < 0) { error(E_NOMEM); return; }
  time = (unsigned long)t /(double)CLOCKS_PER_SEC;
  sprintf(buf, "%s    %.2fs", PRGNAME, time);
  SetWindowText(hmain, buf);    /* display the execution time */
}  /* build() */

/*----------------------------------------------------------------------
  Dialog Functions
----------------------------------------------------------------------*/

static char* fselect (char *title, int load)
{                               /* --- select a file */
  static char fname[_MAX_PATH+1];     /* file name */
  static char path [_MAX_PATH+1];     /* and path */
  static char filter[] =        /* file selection filter */
    "sets of points (*.pts)\0*.pts\0" /* sets of points */
    "all files (*.*)\0*.*\0";         /* all files */
  static OPENFILENAME ofn = {   /* file selector structure */
    sizeof(OPENFILENAME),       /* lStructSize */
    NULL,   NULL,               /* hwndOwner, hInstance */
    filter, NULL,               /* lpstrFilter, lpstrCustomFilter */
    0, 1,                       /* nMaxCustFilter, nFilterIndex */
    fname, _MAX_PATH,           /* lpstrFile, nMaxFile */
    NULL, _MAX_FNAME+_MAX_EXT,  /* lpstrFileTitle, nMaxFileTitle */
    path,                       /* lpstrInitialDir */
    "Select File...",           /* lpstrTitle */
    OFN_PATHMUSTEXIST, 0, 0,    /* Flags, nFileOffset, nFileExtension */
    "nn", 0L,                   /* lpstrDefExt, lCustData */
    NULL, NULL };               /* lpfnHook, lpTemplateName */

  ofn.hwndOwner  = hmain;       /* set window handle */
  ofn.lpstrTitle = title;       /* and title text */
  ofn.Flags = (load) ? OFN_FILEMUSTEXIST : OFN_PATHMUSTEXIST;
  if (GetOpenFileName(&ofn) == 0)
    return NULL;                /* open file selector box */
  return fname;                 /* return name of selected file */
}  /* fselect() */

/*--------------------------------------------------------------------*/

static int load (void)
{                               /* --- File > Load Points */
  char  *fname;                 /* name of file to load */
  int   res;                    /* result of ch_load() */
  CHULL *tmp;                   /* loaded convex hull structure */

  LoadString(hinst, CS_LOAD, text, 80);
  fname = fselect(text, 1);     /* get the dialog box title */
  if (!fname) return -1;        /* and select a file */
  tmp = ch_create();            /* create a convex hull */
  if (!tmp) { error(E_NOMEM); return -1; }
  res = ch_load(tmp, fname);    /* load points of convex hull */
  if (res < 0) {                /* and check for success */
    if (res >= -1) error(E_NOMEM);
    else error((res < -2) ? E_FREAD : E_FOPEN, fname);
    ch_delete(tmp); return -1;  /* on an error delete the created */
  }                             /* convex hull structure and abort */
  if (chull) ch_delete(chull);  /* delete an existing convex hull */
  chull = tmp;                  /* and set the loaded structure */
  state = 0;                    /* and the construction state */
  SetWindowText(hmain, PRGNAME);/* set default window title */
  mn_enable(state);             /* enable/disable menu items */
  return 0;                     /* return 'ok' */
}  /* load() */

/*--------------------------------------------------------------------*/

static void save (void)
{                               /* --- File > Save Points... */
  char *fname;                  /* name of file to save to */

  LoadString(hinst, CS_SAVE, text, 80);
  fname = fselect(text, 0);     /* get the dialog box title */
  if (!fname) return;           /* and select a file */
  if (ch_save(chull, fname) != 0)
    error(E_FWRITE, fname);     /* save points of convex hull */
}  /* save() */

/*----------------------------------------------------------------------
  Dialog Box Procedures
----------------------------------------------------------------------*/

BOOL CALLBACK dp_about (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- 'about' dialog box procedure */
  switch (msg) {                /* evaluate message */
    case WM_COMMAND:            /* --- if command */
      if (wp != IDOK) break;    /* process only 'Ok' button */
    case WM_CLOSE:              /* --- if dialog closed */
      EndDialog(hwnd, 0);       /* terminate dialog */
      return 1;                 /* return 'ok' */
  }
  return 0;                     /* everything else not handled */
}  /* dp_about() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_setup (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- setup dialog procedure */
  int k;                        /* temporary buffer */

  switch (msg) {                /* evaluate message */
    case WM_INITDIALOG:         /* dialog initialization */
      int_set(hwnd, DI_PTCNT, params.ptcnt);  /* number of points */
      LoadString(hinst, CS_CUBE,    text, 80);
      cmb_add(hwnd, DI_REGION, text);
      LoadString(hinst, CS_SPHERE,  text, 80);
      cmb_add(hwnd, DI_REGION, text);
      LoadString(hinst, CS_SURFACE, text, 80);
      cmb_add(hwnd, DI_REGION, text);
      if      (params.region == CH_SPHERE)  k = 1;
      else if (params.region == CH_SURFACE) k = 2;
      else                                  k = 0;
      cmb_set(hwnd, DI_REGION, k);            /* region of points */
      LoadString(hinst, CS_SORTED,  text, 80);
      cmb_add(hwnd, DI_ALG, text);
      LoadString(hinst, CS_CGRAPH,  text, 80);
      cmb_add(hwnd, DI_ALG, text);            /* algorithm */
      cmb_set(hwnd, DI_ALG,   (params.alg   == CH_SORTED) ? 0 : 1);
      LoadString(hinst, CS_WFRAME,  text, 80);
      cmb_add(hwnd, DI_VISIB, text);
      LoadString(hinst, CS_SOLID,   text, 80);
      cmb_add(hwnd, DI_VISIB, text);          /* visibility of egdes */
      cmb_set(hwnd, DI_VISIB, (params.visib == CH_WFRAME) ? 0 : 1);
      break;
    case WM_COMMAND:            /* command */
      switch (wp) {             /* evaluate command code */
        case IDOK:              /* if 'Ok' button pressed */
          params.ptcnt = int_get(hwnd, DI_PTCNT);
          if (params.ptcnt < 4) params.ptcnt = 4;
                                /* number of points */
          switch ((int)cmb_get(hwnd, DI_REGION)) {
            case  1: params.region = CH_SPHERE;  break;
            case  2: params.region = CH_SURFACE; break;
            default: params.region = CH_CUBE;    break;
          }                     /* region of points */
          switch ((int)cmb_get(hwnd, DI_ALG)) {
            case  1: params.alg    = CH_CGRAPH;  break;
            default: params.alg    = CH_SORTED;  break;
          }                     /* construction algorithm */
          switch ((int)cmb_get(hwnd, DI_VISIB)) {
            case  1: params.visib  = CH_SOLID;   break;
            default: params.visib  = CH_WFRAME;  break;
          }                     /* visibility of egdes */
          EndDialog(hwnd, 0);   /* get inputs from edit controls */
          return 1;             /* return 'ok' */
        case IDCANCEL:          /* if 'Cancel' button pressed, */
          EndDialog(hwnd, -1); return 1;   /* terminate dialog */
      } break;
    case WM_CLOSE:              /* dialog box has been closed */
      EndDialog(hwnd, -1); return 1;      /* terminate dialog */
  }
  return 0;                     /* everything else not handled */
}  /* dp_setup() */

/*----------------------------------------------------------------------
  Window Procedure
----------------------------------------------------------------------*/

LRESULT CALLBACK wp_main (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- main window procedure */
  static PAINTSTRUCT ps;        /* for BeginPaint call */
  HDC hdc;                      /* handle of device context */

  switch (msg) {                /* evaluate message */
    case WM_PAINT:              /* --- if to paint window */
      hdc = BeginPaint(hwnd, &ps);
      if (!hdc) return 0;       /* get device context for drawing */
      redraw(hdc);              /* redraw the window contents */
      EndPaint(hwnd, &ps);      /* and release the device context */
      break;

    case WM_COMMAND:            /* --- if menu item selected */
      switch (wp) {             /* evaluate item */
        case MI_LOAD:           /* Menu Item 'Load Map...' */
          if (load() == 0) redraw(NULL);
          return 0;             /* load a map and redraw the window */

        case MI_SAVE:           /* Menu Item 'Save Map...' */
          save(); return 0;     /* save a self-organizing map */

        case MI_QUIT:           /* Menu Item 'Quit' */
          DestroyWindow(hwnd);  /* stop network training */
          return 0;             /* and destroy the window */

        case MI_REINIT:         /* Menu Item 'Reinitialize' */
          init();  redraw(NULL);/* initialize the map */
          return 0;             /* and redraw the window */

        case MI_NEXT:           /* Menu Item 'Next Step' */
          next();  redraw(NULL);/* add next point */
          return 0;             /* and redraw the window */

        case MI_BUILD:          /* Menu Item 'Build all' */
          build(); redraw(NULL);/* build convex hull */
          return 0;             /* and redraw the window */

        case MI_VIEW:           /* Menu Item 'Reset View' */
          p3d_eye  (proj3d, 0,-6, 0); /* init. eye, view, frustrum */
          p3d_view (proj3d, 0, 0, 0);
          p3d_frust(proj3d, M_PI/6, 1e-6, 1e6);
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* invalidate window contents */

        case MI_REDRAW:         /* Menu Item 'Redraw' */
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* invalidate window contents */

        case MI_SETUP:          /* Menu Item 'Setup...' */
          if (DialogBox(hinst, MKIRSC(DB_SETUP), hwnd,
             (DLGPROC)dp_setup) == 0)
            InvalidateRect(hwnd, NULL, 0);
          return 0;             /* setup convex hull construction */

        case MI_ABOUT:          /* Menu Item 'About' */
          DialogBox(hinst, MKIRSC(DB_ABOUT), hwnd, (DLGPROC)dp_about);
          return 0;             /* show dialog box 'About Wsom' */
      } break;

    case WM_LBUTTONDOWN:        /* left   mouse button pressed */
      mode = (wp & MK_SHIFT) ? PITCH : ROTATE;
    case WM_MBUTTONDOWN:        /* middle mouse button pressed */
      if (msg == WM_MBUTTONDOWN) mode = ROLL;
    case WM_RBUTTONDOWN:        /* right  mouse button pressed */
      if (msg == WM_RBUTTONDOWN) mode = (wp & MK_SHIFT) ? YAW : ZOOM;
      mx = LOWORD(lp); my = HIWORD(lp);
      p3d_save(proj3d);         /* save the current mouse position */
      return 0;                 /* and the current projection */

    case WM_LBUTTONUP:          /* mouse button released */
    case WM_MBUTTONUP:
    case WM_RBUTTONUP:
      mode = 0; return 0;       /* clear the movement mode */

    case WM_MOUSEMOVE:          /* if the mouse was moved */
      if (mode <= 0) break;
      mouse(LOWORD(lp), HIWORD(lp));
      InvalidateRect(hwnd, NULL, 0);
      return 0;                /* set new projection and redraw */

    case WM_CLOSE:              /* window closed */
    case WM_DESTROY:            /* or destroyed */
      PostQuitMessage(0);       /* terminate program */
      return 0;
  }
  return DefWindowProc(hwnd, msg, wp, lp);
}  /* wp_main() */

/*----------------------------------------------------------------------
  Main Program
----------------------------------------------------------------------*/

int WINAPI WinMain (HINSTANCE hcurr, HINSTANCE hprev,
                    LPSTR cmdln, int cmd)
{                               /* --- main function */
  WNDCLASS wc;                  /* window class */
  RECT     rc;                  /* window rectangle */
  MSG      msg;                 /* message */

  /* --- create window class --- */
  hinst = hcurr;                /* note instance handle */
  if (!hprev) {                 /* if first instance of this program */
    wc.style         = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc   = wp_main;
    wc.cbClsExtra    = wc.cbWndExtra = 0;
    wc.hInstance     = hinst;
    wc.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = NULL;
    wc.lpszMenuName  = MKIRSC(MN_MAIN);
    wc.lpszClassName = "WsomClass";
    if (!RegisterClass(&wc))    /* register class */
      return 0;                 /* of main window */
  }

  rseed((unsigned)time(NULL));
  proj3d = p3d_create();        /* create a projection object */
  if (!proj3d) error(E_NOMEM);  /* for sphere visualization */
  p3d_eye  (proj3d, 0,-6, 0);   /* initialize eye, view, and frustrum */
  p3d_view (proj3d, 0, 0, 0);
  p3d_frust(proj3d, M_PI/6, 1e-6, 1e6);
  init();                       /* initialize convex hull */

  /* --- get pens and brushes --- */
  pen_black = GetStockObject(BLACK_PEN);
  pen_grey  = CreatePen(PS_SOLID, 1, RGB(0xc0, 0xc0, 0xc0));
  br_white  = GetStockObject(WHITE_BRUSH);

  /* --- create window --- */
  rc.left = 0; rc.right  = INITXEXT -1;
  rc.top  = 0; rc.bottom = INITYEXT -1;
  AdjustWindowRect(&rc, WINSTYLE, 1);
  hmain = CreateWindow("WsomClass", PRGNAME, WINSTYLE,
                       30, 30, rc.right-rc.left, rc.bottom-rc.top,
                       NULL, NULL, hinst, NULL);
  if (!hmain) return 0;         /* create main window */
  hdc_win = GetDC(hmain);
  if (!hdc_win) error(E_INIT);
  hdc_buf = CreateCompatibleDC(hdc_win);
  if (!hdc_buf) error(E_INIT);
  ShowWindow(hmain, cmd);       /* open and */
  UpdateWindow(hmain);          /* update main window */

  /* --- dispatch messages --- */
  while (GetMessage(&msg, NULL, 0, 0) > 0) {
    TranslateMessage(&msg);     /* get messages, translate */
    DispatchMessage (&msg);     /* and dispatch them */
  }

  /* --- clean up --- */
  ReleaseDC(hmain, hdc_win);    /* release device context of window */
  DeleteObject(hdc_buf);        /* delete device context of buffer */
  DeleteObject(buffer);         /* delete drawing buffer */
  DeleteObject(pen_grey);       /* delete created pen */
  return (int)msg.wParam;
}  /* WinMain() */
